from pathlib import Path
import pytest
import gotranx.myokit

here = Path(__file__).parent.absolute()


@pytest.mark.parametrize(
    "cellml_file, num_states, num_parameters",
    [("noble_1962.cellml", 4, 5), ("ToRORd_dynCl_mid.cellml", 45, 112)],
)
def test_cellml_to_gotran_and_back(cellml_file, num_states, num_parameters):
    ode = gotranx.myokit.cellml_to_gotran(
        filename=here / "cellml_files" / cellml_file,
    )
    assert ode.num_states == num_states
    assert ode.num_parameters == num_parameters

    myokit_model = gotranx.myokit.gotran_to_myokit(ode)
    assert myokit_model


@pytest.mark.parametrize(
    "mmt_file, num_states, num_parameters",
    [("example.mmt", 8, 18)],
)
def test_mmt_to_gotran(mmt_file, num_states, num_parameters):
    ode = gotranx.myokit.mmt_to_gotran(
        filename=here / "mmt_files" / mmt_file,
    )
    assert ode.num_states == num_states
    assert ode.num_parameters == num_parameters



from __future__ import annotations
from textwrap import dedent, indent
from structlog import get_logger

logger = get_logger()


def init_state_values(name, state_names, state_values, code):
    indented_code = indent(code, "    ")
    indent_values = indent(", ".join(f"{n}={v}" for n, v in zip(state_names, state_values)), "    ")
    return dedent(
        f"""
void init_state_values(double* {name}){{
    /*
{indent_values}
    */
{indented_code}
}}
""",
    )


def init_parameter_values(name, parameter_names, parameter_values, code):
    indented_code = indent(code, "    ")
    indent_values = indent(
        ", ".join(f"{n}={v}" for n, v in zip(parameter_names, parameter_values)), "    "
    )
    return dedent(
        f"""
void init_parameter_values(double* {name}){{
    /*
{indent_values}
    */
{indented_code}
}}
""",
    )


def method(
    name: str,
    args: str,
    states: str,
    parameters: str,
    values: str,
    return_name: None = None,
    num_return_values: int = 0,
    **kwargs,
):
    indent_states = indent(states, "    ")
    indent_parameters = indent(parameters, "    ")
    indent_values = indent(values, "    ")
    return dedent(
        f"""
void {name}({args}){{

    // Assign states
{indent_states}

    // Assign parameters
{indent_parameters}

    // Assign expressions
{indent_values}
}}
""",
    )


def method_index(data: dict[str, int], method_name) -> str:
    logger.debug(f"Generating {method_name}_index with {len(data)} values")
    local_template = dedent(
        """
    {if_stm} (strcmp(name, "{name}") == 0) {{
        return {index};
    }}
    """
    )
    code = []
    code.append(f"// {method_name.capitalize()} index")
    code.append(f"int {method_name}_index(const char name[])")
    code.append("{")
    for i, (name, index) in enumerate(data.items()):
        if_stm = "if" if i == 0 else "else if"
        code.append(indent(local_template.format(if_stm=if_stm, name=name, index=index), "    "))
    code.append(indent("return -1;", "    "))
    code.append("}")

    return "\n".join(code)


def parameter_index(data: dict[str, int]) -> str:
    return method_index(data, "parameter")


def state_index(data: dict[str, int]) -> str:
    return method_index(data, "state")


def monitor_index(data: dict[str, int]) -> str:
    return method_index(data, "monitor")


def missing_index(data: dict[str, int]) -> str:
    return method_index(data, "monitor")



# flake8: noqa: E501
import sys
from unittest import mock

import pytest
from gotranx.schemes import get_scheme
from gotranx.codegen import CCodeGenerator
from gotranx.codegen import RHSArgument
from gotranx.ode import make_ode


@pytest.fixture(scope="module")
def ode(trans, parser):
    expr = """
    parameters(a=0)
    parameters("My component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states("My component", x=1.0, y=2.0,z=3.05)

    expressions("My component")
    dy_dt = x*(rho - z) - y # millivolt
    dx_dt = sigma*(-x + y)
    dz_dt = -beta*z + x*y
    """
    tree = parser.parse(expr)
    return make_ode(*trans.transform(tree), name="lorentz")


@pytest.fixture(scope="module")
def codegen(ode) -> CCodeGenerator:
    return CCodeGenerator(ode)


def test_c_codegen_initial_state_values(codegen: CCodeGenerator):
    assert codegen.initial_state_values() == (
        "\nvoid init_state_values(double *states)\n"
        "{\n"
        "    /*\n"
        "    x=1.0, y=2.0, z=3.05\n"
        "    */\n"
        "    states[0] = 1.0;\n"
        "    states[1] = 2.0;\n"
        "    states[2] = 3.05;\n"
        "}\n"
    )


def test_c_codegen_initial_parameter_values(codegen: CCodeGenerator):
    assert codegen.initial_parameter_values() == (
        "\nvoid init_parameter_values(double *parameters)\n"
        "{\n"
        "    /*\n"
        "    a=0, beta=2.4, rho=21.0, sigma=12.0\n"
        "    */\n"
        "    parameters[0] = 0;\n"
        "    parameters[1] = 2.4;\n"
        "    parameters[2] = 21.0;\n"
        "    parameters[3] = 12.0;\n"
        "}\n"
    )


def test_c_codegen_initial_parameter_values_no_clang_format(ode):
    with mock.patch.dict(sys.modules, {"clang_format_docs": None}):
        codegen = CCodeGenerator(ode)
        assert codegen.initial_parameter_values() == (
            "\nvoid init_parameter_values(double* parameters){\n"
            "    /*\n"
            "    a=0, beta=2.4, rho=21.0, sigma=12.0\n"
            "    */\n"
            "    parameters[0] = 0;\n"
            "    parameters[1] = 2.4;\n"
            "    parameters[2] = 21.0;\n"
            "    parameters[3] = 12.0;\n"
            "}\n"
        )


@pytest.mark.parametrize(
    "order, arguments",
    [
        (
            RHSArgument.stp,
            (
                "const double *__restrict states, const double t, "
                "const double *__restrict parameters, double *values"
            ),
        ),
        (
            RHSArgument.spt,
            (
                "const double *__restrict states, "
                "const double *__restrict parameters, const double t, double *values"
            ),
        ),
        (
            RHSArgument.tsp,
            (
                "const double t, const double *__restrict states, "
                "const double *__restrict parameters, double *values"
            ),
        ),
    ],
)
def test_c_codegen_rhs(order: str, arguments: str, codegen: CCodeGenerator):
    assert codegen.rhs(order=order) == (
        f"\nvoid rhs({arguments})\n"
        "{\n"
        "\n"
        "    // Assign states\n"
        "    const double x = states[0];\n"
        "    const double y = states[1];\n"
        "    const double z = states[2];\n"
        "\n"
        "    // Assign parameters\n"
        "    const double a = parameters[0];\n"
        "    const double beta = parameters[1];\n"
        "    const double rho = parameters[2];\n"
        "    const double sigma = parameters[3];\n"
        "\n"
        "    // Assign expressions\n"
        "    const double dx_dt = sigma * (-x + y);\n"
        "    values[0] = dx_dt;\n"
        "    const double dy_dt = x * (rho - z) - y;\n"
        "    values[1] = dy_dt;\n"
        "    const double dz_dt = (-beta) * z + x * y;\n"
        "    values[2] = dz_dt;\n"
        "}\n"
    )


def test_c_codegen_forward_euler(codegen: CCodeGenerator):
    assert codegen.scheme(get_scheme("forward_euler")) == (
        "\nvoid forward_euler(const double *__restrict states, const double t, const double dt,"
        "\n                   const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double x = states[0];"
        "\n    const double y = states[1];"
        "\n    const double z = states[2];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    values[0] = dt * dx_dt + x;"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    values[1] = dt * dy_dt + y;"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    values[2] = dt * dz_dt + z;"
        "\n}"
        "\n"
    )


def test_c_codegen_forward_generalized_rush_larsen(codegen: CCodeGenerator):
    assert codegen.scheme(get_scheme("forward_generalized_rush_larsen")) == (
        "\nvoid forward_generalized_rush_larsen(const double *__restrict states, const double t, const double dt,"
        "\n                                     const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double x = states[0];"
        "\n    const double y = states[1];"
        "\n    const double z = states[2];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    const double dx_dt_linearized = -sigma;"
        "\n    values[0] = x + ((fabs(dx_dt_linearized) > 1e-08) ? (dx_dt * (exp(dt * dx_dt_linearized) - 1) / dx_dt_linearized)"
        "\n                                                      : (dt * dx_dt));"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    const double dy_dt_linearized = -1;"
        "\n    values[1] = y + ((fabs(dy_dt_linearized) > 1e-08) ? (dy_dt * (exp(dt * dy_dt_linearized) - 1) / dy_dt_linearized)"
        "\n                                                      : (dt * dy_dt));"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    const double dz_dt_linearized = -beta;"
        "\n    values[2] = z + ((fabs(dz_dt_linearized) > 1e-08) ? (dz_dt * (exp(dt * dz_dt_linearized) - 1) / dz_dt_linearized)"
        "\n                                                      : (dt * dz_dt));"
        "\n}"
        "\n"
    )


def test_c_remove_unused_rhs(ode_unused):
    codegen_orig = CCodeGenerator(ode_unused)
    assert codegen_orig.rhs() == (
        "\nvoid rhs(const double t, const double *__restrict states, const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double unused_state = states[0];"
        "\n    const double x = states[1];"
        "\n    const double y = states[2];"
        "\n    const double z = states[3];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n    const double unused_parameter = parameters[4];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double unused_expression = 0;"
        "\n    const double dunused_state_dt = 0;"
        "\n    values[0] = dunused_state_dt;"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    values[1] = dx_dt;"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    values[2] = dy_dt;"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    values[3] = dz_dt;"
        "\n}"
        "\n"
    )
    codegen_remove = CCodeGenerator(ode_unused, remove_unused=True)
    assert codegen_remove.rhs() == (
        "\nvoid rhs(const double t, const double *__restrict states, const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double x = states[1];"
        "\n    const double y = states[2];"
        "\n    const double z = states[3];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dunused_state_dt = 0;"
        "\n    values[0] = dunused_state_dt;"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    values[1] = dx_dt;"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    values[2] = dy_dt;"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    values[3] = dz_dt;"
        "\n}"
        "\n"
    )


def test_c_remove_unused_forward_explicit_euler(ode_unused):
    codegen_orig = CCodeGenerator(ode_unused)
    assert codegen_orig.scheme(get_scheme("forward_euler")) == (
        "\nvoid forward_euler(const double *__restrict states, const double t, const double dt,"
        "\n                   const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double unused_state = states[0];"
        "\n    const double x = states[1];"
        "\n    const double y = states[2];"
        "\n    const double z = states[3];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n    const double unused_parameter = parameters[4];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double unused_expression = 0;"
        "\n    const double dunused_state_dt = 0;"
        "\n    values[0] = dt * dunused_state_dt + unused_state;"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    values[1] = dt * dx_dt + x;"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    values[2] = dt * dy_dt + y;"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    values[3] = dt * dz_dt + z;"
        "\n}"
        "\n"
    )
    codegen_remove = CCodeGenerator(ode_unused, remove_unused=True)
    assert codegen_remove.scheme(get_scheme("forward_euler")) == (
        "\nvoid forward_euler(const double *__restrict states, const double t, const double dt,"
        "\n                   const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double unused_state = states[0];"
        "\n    const double x = states[1];"
        "\n    const double y = states[2];"
        "\n    const double z = states[3];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dunused_state_dt = 0;"
        "\n    values[0] = dt * dunused_state_dt + unused_state;"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    values[1] = dt * dx_dt + x;"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    values[2] = dt * dy_dt + y;"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    values[3] = dt * dz_dt + z;"
        "\n}"
        "\n"
    )


def test_c_remove_unused_forward_generalized_rush_larsen(ode_unused):
    codegen_orig = CCodeGenerator(ode_unused)
    assert codegen_orig.scheme(get_scheme("forward_generalized_rush_larsen")) == (
        "\nvoid forward_generalized_rush_larsen(const double *__restrict states, const double t, const double dt,"
        "\n                                     const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double unused_state = states[0];"
        "\n    const double x = states[1];"
        "\n    const double y = states[2];"
        "\n    const double z = states[3];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n    const double unused_parameter = parameters[4];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double unused_expression = 0;"
        "\n    const double dunused_state_dt = 0;"
        "\n    values[0] = dt * dunused_state_dt + unused_state;"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    const double dx_dt_linearized = -sigma;"
        "\n    values[1] = x + ((fabs(dx_dt_linearized) > 1e-08) ? (dx_dt * (exp(dt * dx_dt_linearized) - 1) / dx_dt_linearized)"
        "\n                                                      : (dt * dx_dt));"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    const double dy_dt_linearized = -1;"
        "\n    values[2] = y + ((fabs(dy_dt_linearized) > 1e-08) ? (dy_dt * (exp(dt * dy_dt_linearized) - 1) / dy_dt_linearized)"
        "\n                                                      : (dt * dy_dt));"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    const double dz_dt_linearized = -beta;"
        "\n    values[3] = z + ((fabs(dz_dt_linearized) > 1e-08) ? (dz_dt * (exp(dt * dz_dt_linearized) - 1) / dz_dt_linearized)"
        "\n                                                      : (dt * dz_dt));"
        "\n}"
        "\n"
    )
    codegen_remove = CCodeGenerator(ode_unused, remove_unused=True)
    assert codegen_remove.scheme(get_scheme("forward_generalized_rush_larsen")) == (
        "\nvoid forward_generalized_rush_larsen(const double *__restrict states, const double t, const double dt,"
        "\n                                     const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double unused_state = states[0];"
        "\n    const double x = states[1];"
        "\n    const double y = states[2];"
        "\n    const double z = states[3];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dunused_state_dt = 0;"
        "\n    values[0] = dt * dunused_state_dt + unused_state;"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    const double dx_dt_linearized = -sigma;"
        "\n    values[1] = x + ((fabs(dx_dt_linearized) > 1e-08) ? (dx_dt * (exp(dt * dx_dt_linearized) - 1) / dx_dt_linearized)"
        "\n                                                      : (dt * dx_dt));"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    const double dy_dt_linearized = -1;"
        "\n    values[2] = y + ((fabs(dy_dt_linearized) > 1e-08) ? (dy_dt * (exp(dt * dy_dt_linearized) - 1) / dy_dt_linearized)"
        "\n                                                      : (dt * dy_dt));"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    const double dz_dt_linearized = -beta;"
        "\n    values[3] = z + ((fabs(dz_dt_linearized) > 1e-08) ? (dz_dt * (exp(dt * dz_dt_linearized) - 1) / dz_dt_linearized)"
        "\n                                                      : (dt * dz_dt));"
        "\n}"
        "\n"
    )


def test_c_conditional_expression_assignment(parser, trans):
    expr = """
    states(v=0)
    ah = Conditional(Ge(v, -40), 0, 0.057*exp(-(v + 80)/6.8))
    dv_dt = 0
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    ode = make_ode(*result, name="conditional")
    codegen = CCodeGenerator(ode)
    assert codegen.rhs() == (
        "\nvoid rhs(const double t, const double *__restrict states, const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double v = states[0];"
        "\n"
        "\n    // Assign parameters"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dv_dt = 0;"
        "\n    values[0] = dv_dt;"
        "\n    const double ah = (v >= -40) ? 0 : 0.057 * exp((-(v + 80)) / 6.8);"
        "\n}"
        "\n"
    )


def test_c_conditional_expression_advanced(parser, trans):
    expr = """
    states(v=0.5)
    parameters(g=1)
    gammasu  = g*Conditional(Gt(Gt(v,0)*v, Lt(v, -1)*(-v-1)), Gt(v,0)*v, Lt(v, -1)*(-v-1))
    dv_dt = 0
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    ode = make_ode(*result, name="conditional")
    codegen = CCodeGenerator(ode)
    assert codegen.rhs() == (
        "\nvoid rhs(const double t, const double *__restrict states, const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double v = states[0];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double g = parameters[0];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dv_dt = 0;"
        "\n    values[0] = dv_dt;"
        "\n    const double gammasu ="
        "\n        g * ((((v > 0 && v < -1) ? (v > -v - 1) : (((v > 0) ? (v > 0) : (((v < -1) ? (v > -1) : (0)))))))"
        "\n                 ? (v * ((v > 0) ? (1) : (0)))"
        "\n                 : ((-v - 1) * ((v < -1) ? (1) : (0))));"
        "\n}"
        "\n"
    )


def test_c_codegen_parameter_index(codegen: CCodeGenerator):
    assert codegen.parameter_index() == (
        "// Parameter index"
        "\nint parameter_index(const char name[])"
        "\n{"
        "\n"
        '\n    if (strcmp(name, "a") == 0)'
        "\n    {"
        "\n        return 0;"
        "\n    }"
        "\n"
        '\n    else if (strcmp(name, "beta") == 0)'
        "\n    {"
        "\n        return 1;"
        "\n    }"
        "\n"
        '\n    else if (strcmp(name, "rho") == 0)'
        "\n    {"
        "\n        return 2;"
        "\n    }"
        "\n"
        '\n    else if (strcmp(name, "sigma") == 0)'
        "\n    {"
        "\n        return 3;"
        "\n    }"
        "\n"
        "\n    return -1;"
        "\n}"
    )


def test_c_codegen_state_index(codegen: CCodeGenerator):
    assert codegen.state_index() == (
        "// State index"
        "\nint state_index(const char name[])"
        "\n{"
        "\n"
        '\n    if (strcmp(name, "x") == 0)'
        "\n    {"
        "\n        return 0;"
        "\n    }"
        "\n"
        '\n    else if (strcmp(name, "y") == 0)'
        "\n    {"
        "\n        return 1;"
        "\n    }"
        "\n"
        '\n    else if (strcmp(name, "z") == 0)'
        "\n    {"
        "\n        return 2;"
        "\n    }"
        "\n"
        "\n    return -1;"
        "\n}"
    )


def test_c_codegen_monitor_index(codegen: CCodeGenerator):
    assert codegen.monitor_index() == (
        "// Monitor index"
        "\nint monitor_index(const char name[])"
        "\n{"
        "\n"
        '\n    if (strcmp(name, "dx_dt") == 0)'
        "\n    {"
        "\n        return 0;"
        "\n    }"
        "\n"
        '\n    else if (strcmp(name, "dy_dt") == 0)'
        "\n    {"
        "\n        return 1;"
        "\n    }"
        "\n"
        '\n    else if (strcmp(name, "dz_dt") == 0)'
        "\n    {"
        "\n        return 2;"
        "\n    }"
        "\n"
        "\n    return -1;"
        "\n}"
    )


def test_c_codegen_monitor(codegen: CCodeGenerator):
    assert codegen.monitor_values() == (
        "\nvoid monitor_values(const double t, const double *__restrict states, const double *__restrict parameters,"
        "\n                    double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double x = states[0];"
        "\n    const double y = states[1];"
        "\n    const double z = states[2];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    values[0] = dx_dt;"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    values[1] = dy_dt;"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    values[2] = dz_dt;"
        "\n}"
        "\n"
    )



from __future__ import annotations
from collections import defaultdict
import warnings
from pathlib import Path
from typing import overload

import sympy as sp
import myokit
import myokit.formats.sympy
import myokit.lib.guess
import myokit.formats.cellml

from . import atoms
from .ode_component import MyokitComponent
from .ode import ODE

reserved_names = {name for name in dir(sp) if not name.startswith("_")}


@overload
def extract_unit(unit: str) -> str: ...


@overload
def extract_unit(unit: None) -> None: ...


def extract_unit(unit: str | None) -> str | None:
    """Extract unit from myokit unit

    Parameters
    ----------
    unit : str | None
        The myokit unit

    Returns
    -------
    str | None
        Return the unit as a string
    """
    if unit is None:
        return None
    return str(unit).strip("[").strip("]").replace("^", "**")


def extract_nested_variables(
    model: myokit.Model,
) -> tuple[dict[sp.Symbol, sp.Symbol], dict[str, dict[sp.Symbol, sp.Symbol]]]:
    """Extract nested variables from myokit model

    Parameters
    ----------
    model : myokit.Model
        The myokit model

    Returns
    -------
    tuple[dict[sp.Symbol, sp.Symbol], dict[str, dict[sp.Symbol, sp.Symbol]]]
        A tuple containing all substitutions and component substitutions
    """

    all_subs: dict[sp.Symbol, sp.Symbol] = {}
    component_subs: dict[str, dict[sp.Symbol, sp.Symbol]] = defaultdict(dict)

    def f(component, all_subs, component_subs):
        component_subs_ = {}
        for var in component.variables():
            name = var.uname()
            if name in reserved_names:
                name = f"{name}_"

            component_subs[component.name()][sp.Symbol(var.name())] = sp.Symbol(name)
            all_subs[sp.Symbol(var.qname())] = sp.Symbol(name)
            all_subs, component_subs_ = f(var, all_subs, component_subs)
        component_subs.update(component_subs_)
        return all_subs, component_subs

    for component in model.components():
        all_subs, component_subs = f(component, all_subs, component_subs)

    return all_subs, component_subs


def mmt_to_gotran(filename: str | Path) -> ODE:
    """Convert a myokit model to gotran ODE

    Parameters
    ----------
    filename : str | Path
        The filename of the myokit model

    Returns
    -------
    gotranx.ode.ODE
        The gotran ODE
    """
    model, protocol, _ = myokit.load(filename)

    return myokit_to_gotran(model, protocol=protocol)


def myokit_to_gotran(model: myokit.Model, protocol=None) -> ODE:
    """Convert a myokit model to gotran ODE

    Parameters
    ----------
    model : myokit.Model
        The myokit model
    protocol : _type_, optional
        The stimulus protocol used in myokit, by default None

    Returns
    -------
    gotranx.ode.ODE
        The gotran ODE
    """
    # Embed protocol
    if protocol is not None:
        model = model.clone()
        if not myokit.lib.guess.add_embedded_protocol(model, protocol):
            warnings.warn("Unable to embed stimulus protocol.")

    model.create_unique_names()

    all_subs, component_subs = extract_nested_variables(model)

    initial_values = model.initial_values()
    components = []
    for component in model.components():
        states = []
        parameters = []
        intermediates = []
        derivatives = []
        for var in component.variables(deep=True):
            name = var.uname()
            if name in reserved_names:
                name = f"{name}_"

            if name == "time":
                # Skip time variable
                continue

            if var.is_state():
                state = atoms.State(
                    name=name,
                    value=initial_values[var.index()],
                    components=(component.name(),),
                    unit_str=extract_unit(var.unit()),
                    description=var.meta.get("desc", ""),
                )
                states.append(state)
                with sp.core.parameters.evaluate(False):
                    expr = myokit.formats.sympy.write(var.eq().rhs)
                    expr = expr.xreplace({v.name(): v.uname() for v in var.variables(deep=True)})
                    expr = expr.xreplace(component_subs.get(component.name(), {}))
                    expr = expr.xreplace(all_subs)

                state_der = atoms.StateDerivative(
                    name=f"d{name}_dt",
                    expr=expr,
                    value=None,
                    components=(component.name(),),
                    unit_str=extract_unit(var.unit()),
                    description=var.meta.get("desc", ""),
                    state=state,
                )
                derivatives.append(state_der)

            else:
                with sp.core.parameters.evaluate(False):
                    expr = myokit.formats.sympy.write(var.rhs())
                if expr.is_Number:
                    parameter = atoms.Parameter(
                        name=name,
                        value=var.value(),
                        components=(component.name(),),
                        unit_str=extract_unit(var.unit()),
                        description=var.meta.get("desc", ""),
                    )
                    parameters.append(parameter)

                else:
                    with sp.core.parameters.evaluate(False):
                        expr = expr.xreplace(
                            {v.name(): v.uname() for v in var.variables(deep=True)}
                        )
                        expr = expr.xreplace(component_subs.get(component.name(), {}))
                        expr = expr.xreplace(all_subs)

                    intermediate = atoms.Intermediate(
                        name=name,
                        expr=expr,
                        value=None,
                        components=(component.name(),),
                        unit_str=extract_unit(var.unit()),
                        description=var.meta.get("desc", ""),
                    )
                    intermediates.append(intermediate)

        comp = MyokitComponent(
            name=component.name(),
            states=frozenset(states),
            parameters=frozenset(parameters),
            intermediates=frozenset(intermediates),
            state_derivatives=frozenset(derivatives),
        )
        components.append(comp)

    return ODE(
        components=components,
        name=model.name(),
        comments=(atoms.Comment(model.meta.get("desc", "")),),
    )


def gotran_to_myokit(ode: ODE) -> myokit.Model:
    """Convert a gotran ODE to myokit model

    Parameters
    ----------
    ode : gotranx.ode.ODE
        The gotran ODE

    Returns
    -------
    myokit.Model
        The myokit model
    """
    model = myokit.Model(ode.name)
    model.meta["author"] = "GotranX API"
    comp = model.add_component("engine")
    # breakpoint()
    time = comp.add_variable("time", binding="time")
    time.set_rhs(0)
    model._bindings["time"] = time
    # model._register_binding("time")

    # First we need to add all variables to the model
    global_var_map = {sp.Symbol("time"): sp.Symbol("engine.time")}
    for component in ode.components:
        comp = model.add_component(component.name)
        for state_derivative in component.state_derivatives:
            state = state_derivative.state
            var = comp.add_variable(state.name)
            global_var_map[sp.Symbol(state.name)] = sp.Symbol(var.qname())

        for parameter in component.parameters:
            var = comp.add_variable(parameter.name)
            var.set_rhs(parameter.value)
            global_var_map[sp.Symbol(parameter.name)] = sp.Symbol(var.qname())

        for intermediate in component.intermediates:
            var = comp.add_variable(intermediate.name)
            global_var_map[sp.Symbol(intermediate.name)] = sp.Symbol(var.qname())

    sympy_reader = myokit.formats.sympy.SymPyExpressionReader(model=model)
    # Then we can add expressions
    for component in ode.components:
        comp = model[component.name]

        for state_derivative in component.state_derivatives:
            state = state_derivative.state
            v = comp[state.name]

            expr = state_derivative.expr.xreplace(global_var_map)
            expr = sympy_reader.ex(expr)
            v.set_rhs(expr)

            v.promote(state.value)

        for intermediate in component.intermediates:
            v = comp[intermediate.name]
            expr = intermediate.expr.xreplace(global_var_map)
            expr = sympy_reader.ex(expr)
            v.set_rhs(expr)

    model.validate()
    return model


def cellml_to_gotran(filename: str | Path) -> ODE:
    """Convert a cellml file to gotran ODE by via myokit


    Parameters
    ----------
    filename : str | Path
        The filename of the cellml file

    Returns
    -------
    gotranx.ode.ODE
        The gotran ODE
    """
    myokit_model = myokit.formats.cellml.CellMLImporter().model(filename)
    return myokit_to_gotran(myokit_model)


def gotran_to_cellml(ode: ODE, filename: str | Path) -> None:
    """Convert a gotran ODE to cellml file via myokit

    Parameters
    ----------
    ode : gotranx.ode.ODE
        The gotran ODE
    filename : str | Path
        The filename of the cellml file
    """
    model = gotran_to_myokit(ode)
    myokit.formats.cellml.CellMLExporter().model(filename, model, version="1.0")



from __future__ import annotations
import typing
from enum import Enum
from sympy.printing.pycode import PythonCodePrinter

# from sympy.printing.numpy import NumPyPrinter
from sympy.codegen.ast import Assignment
import sympy
import structlog
from functools import partial

from ..ode import ODE
from .. import templates
from .base import CodeGenerator, Func, RHSArgument, SchemeArgument, _print_Piecewise

logger = structlog.get_logger()


class Format(str, Enum):
    black = "black"
    ruff = "ruff"
    none = "none"


# class GotranPythonCodePrinter(NumPyPrinter):
class GotranPythonCodePrinter(PythonCodePrinter):
    _kf = {k: f"numpy.{v.replace('math.', '')}" for k, v in PythonCodePrinter._kf.items()}
    _kc = {k: f"numpy.{v.replace('math.', '')}" for k, v in PythonCodePrinter._kc.items()}

    def _hprint_Pow(self, expr, rational=False, sqrt="numpy.sqrt"):
        return super()._hprint_Pow(expr, rational, sqrt)

    def _print_MatrixElement(self, expr):
        if expr.parent.shape[1] == 1:
            # Then this is a column vector
            return f"{self._print(expr.parent)}[{expr.i}]"
        elif expr.parent.shape[0] == 1:
            # Then this is a row vector
            return f"{self._print(expr.parent)}[{expr.j}]"
        else:
            return super()._print_MatrixElement(expr)

    def _print_Float(self, flt):
        return self._print(str(float(flt)))

    def _print_Piecewise(self, expr):
        result = []

        if isinstance(expr.args[0][0], Assignment):
            lhs = super()._print(expr.args[0][0].lhs)
            result.append(f"{super()._print(lhs)} = ")
            all_lsh_equal = True
            for arg in expr.args:
                result.append("numpy.where(")
                result.append(f"{super()._print(arg[1])}")
                result.append(", ")
                result.append(f"{super()._print(arg[0].rhs)}")
                result.append(", ")
                all_lsh_equal = all_lsh_equal and super()._print(arg[0].lhs) == lhs

            assert all_lsh_equal, "All assignments in Piecewise must have the same lhs"

            if super()._print(arg[1]) == "True":
                result = result[:-6]
                result.append(f", {super()._print(arg[0].rhs)}")
            else:
                raise ValueError("Last condition in Piecewise must be True")

            result.append(")" * (len(expr.args) - 1))

        else:
            conds, exprs = _print_Piecewise(self, expr)

            for c, e in zip(conds, exprs):
                result.append("numpy.where(")
                result.append(f"{c}")
                result.append(", ")
                result.append(f"{e}")
                result.append(", ")

            if c == "True":
                result = result[:-6]
                result.append(f", {e}")
            else:
                raise ValueError("Last condition in Piecewise must be True")

            result.append(")" * (len(conds) - 1))

        return "".join(result)

    def _print_And(self, expr):
        if len(expr.args) == 2:
            value = f"numpy.logical_and({self._print(expr.args[0])}, {self._print(expr.args[1])})"
        else:
            args = ", ".join(self._print(arg) for arg in expr.args)
            value = f"numpy.logical_and.reduce(({args}))"

        return value

    def _print_Or(self, expr):
        # value = super()._print_Or(expr)
        if len(expr.args) == 2:
            value = f"numpy.logical_or({self._print(expr.args[0])}, {self._print(expr.args[1])})"
        else:
            args = ", ".join(self._print(arg) for arg in expr.args)
            value = f"numpy.logical_or.reduce(({args}))"

        return value


def get_formatter(format: Format) -> typing.Callable[[str], str]:
    if format == Format.none:
        return lambda x: x

    elif format == Format.black:
        try:
            import black
        except ImportError:
            logger.warning("Cannot apply black, please install 'black'")
            return lambda x: x
        else:
            return partial(black.format_str, mode=black.Mode())

    elif format == Format.ruff:
        try:
            import ruff.__main__

            ruff_bin = ruff.__main__.find_ruff_bin()
        except ImportError:
            logger.warning("Cannot apply ruff, please install 'ruff'")
            return lambda x: x
        else:
            import subprocess

            def func(code: str) -> str:
                return subprocess.check_output(
                    [ruff_bin, "format", "-"], input=code, encoding="utf-8"
                )

            return func

    else:
        raise ValueError(f"Unknown format {format}")


class PythonCodeGenerator(CodeGenerator):
    def __init__(self, ode: ODE, format: Format = Format.black, *args, **kwargs) -> None:
        super().__init__(ode, *args, **kwargs)

        self._printer = GotranPythonCodePrinter()

        setattr(self, "_formatter", get_formatter(format=format))

    @property
    def printer(self):
        return self._printer

    @property
    def template(self):
        return templates.python

    def imports(self) -> str:
        return self._format("import numpy")

    def _rhs_arguments(
        self,
        order: RHSArgument | str = RHSArgument.stp,
    ) -> Func:
        value = RHSArgument.get_value(order)

        argument_dict = {
            "s": "states",
            "t": "t",
            "p": "parameters",
        }

        argument_list = [argument_dict[v] for v in value]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
            return_name="values",
            num_return_values=self.ode.num_states,
            values_type="numpy.zeros_like(states, dtype=numpy.float64)",
        )

    def _scheme_arguments(
        self,
        order: SchemeArgument | str = SchemeArgument.stdp,
    ) -> Func:
        value = SchemeArgument.get_value(order)

        argument_dict = {
            "s": "states",
            "t": "t",
            "d": "dt",
            "p": "parameters",
        }

        argument_list = [argument_dict[v] for v in value]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
            return_name="values",
            num_return_values=self.ode.num_states,
            values_type="numpy.zeros_like(states, dtype=numpy.float64)",
        )



from pathlib import Path

import pytest
from gotranx.load import load_ode, ode_from_string
from gotranx.exceptions import ODEFileNotFound, InvalidODEException, ComponentNotCompleteError


@pytest.fixture
def path():
    _path = Path("tmp_ode_file.ode")
    yield _path
    _path.unlink()


def test_load_nonexisting_ode_raises_ODEFileNotFound():
    with pytest.raises(ODEFileNotFound):
        load_ode("invalid_ode.ode")


def test_load_ode(path):
    path.write_text(
        """
    states("First component", "X-gate", x = 1, xr=3.14)
    states("First component", "Y-gate", y = 1)
    states("Second component", z=1)
    parameters("First component", a=1, b=2)
    parameters("Second component", c=3)

    expressions("First component")
    d = a + b * 2 - 3 / c

    expressions("First component", "X-gate")
    dx_dt=a+1
    dxr_dt = (x / b) - d * xr

    expressions("First component", "Y-gate")
    dy_dt = 2 * d - 1

    expressions("Second component")
    dz_dt = 1 + x - y
    """,
    )

    ode = load_ode(path)
    assert ode.num_components == 4
    assert set([c.name for c in ode.components]) == {
        "First component",
        "Second component",
        "X-gate",
        "Y-gate",
    }
    assert ode.num_states == 4
    assert ode.num_parameters == 3


def test_load_invalid_ode():
    with pytest.raises(InvalidODEException):
        ode_from_string("states(x=1)")


def test_load_incomplete_ode():
    with pytest.raises(ComponentNotCompleteError):
        ode_from_string("states(x=1, y=2)\ndx_dt = x + y")



from pathlib import Path

from lark import Lark


_here = Path(__file__).absolute().parent


def load_grammar() -> str:
    """Load the grammar

    Returns
    -------
    str
        The grammar
    """
    with open(_here / "ode.lark", "r") as f:
        text = f.read()
    return text


class Parser(Lark):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(load_grammar(), *args, **kwargs)
